#!/bin/perl

######################################################################
# indexer.pl - The Perlfect Indexer                                  #
# version 2.0                                                        #
#                                                                    #
# (c) 1998 Giorgos Zervas & Nick Moraitakis,  Perlfect Solutions     #
# giorgos@perlfect.com                                               #
# http://perlfect.com                                                #
#                                                                    #
# 12 November 1999                                                   #
#                                                                    #
# requires the following perl modules:                               #
# Data::Dumper                                                       #
#                                                                    #
# suggested scripts:                                                 #
# search.pl >=v1.0                                                   #
######################################################################

use Data::Dumper;
use CGI;
use CGI::Carp qw(fatalsToBrowser);

require 'search_conf.pl';

$rcs = ' $Id: indexer.pl,v 1.5 1999/04/16 16:10:20 perlfect Exp $ ' ;	

my $tab = " " x 4;

my @exclude_dirs;
my @exclude_files;

my %index;
my %freq;
my %file_index;
my @stopwords;

my $f_index = 0;
$MIN_WORD_LENGTH--;

if (-t STDOUT) {
  print "Perlfect Indexer $VERSION\n\n";
} else {
  my $query = new CGI;
  print $query->header;
  print "<BODY bgcolor=\"#ffffff\">\n";
  print "<H2>Perlfect Indexer $VERSION</H2>\n";
}

if ($DEBUG) {
  if (-t STDOUT) {
    print "WARNING: You are running the indexer in debug mode. This option is only useful to developers. To disable debug mode open the indexer.pl script and set the \$DEBUG variable to 0(zero).\n"; 
  } else {
    print "<P><FONT color=\"#990000\"><B>WARNING:</B></FONT> You are running the indexer in debug mode. This option is only usefule to developers. To disable debug mode open the indexer.pl script and set the \$DEBUG variable to 0(zero).<P>\n"; 
  }
}
load_exclude_dirs();
load_exclude_files();
load_stopwords();
find_valid_files($BASE_DIR, 0);
create_index();
flatten_index();

open(FILEINDEX, ">$FILE_INDEX_FILE") or die "Cannot open $FILE_INDEX_FILE";
print FILEINDEX Data::Dumper->Dump([\%file_index],[qw(rfile_index)]);
close(FILEINDEX);
chmod(oct(755), $FILE_INDEX_FILE);

if (-t STDOUT) {
  print "File index stored in :$FILE_INDEX_FILE\n";
} else {
  print "<P>File index stored in :<B>$FILE_INDEX_FILE</B>\n";
}

if ($DEBUG) {
  open(INDEX, ">$INDEX_FILE") or die "Cannot open $INDEX_FILE";
  print INDEX Dumper(\%index);
  close(INDEX);
  chmod(oct(755), $INDEX_FILE);

  if (-t STDOUT) {
    print "Expanded index stored in $INDEX_FILE\n";
  } else {
    print "<P>Expanded index stored in: <B>$INDEX_FILE</B>\n";
  }

  $Data::Dumper::Indent=3;  #pretty printing of data structures
} else {
  $Data::Dumper::Indent=0;  #compact printing of data structures
}

open(FLATINDEX, ">$FLAT_INDEX_FILE") or die "Cannot open $FLAT_INDEX_FILE";
print FLATINDEX Data::Dumper->Dump([\%freq],[qw(rfreq)]);
close(FLATINDEX);
chmod(oct(755), $FLAT_INDEX_FILE);

if (-t STDOUT) {
  print "Word(flat) index stored in $FLAT_INDEX_FILE\n";
} else {
  print "<P>Word(flat) index stored in: <B>$FLAT_INDEX_FILE</B>\n";
}

sub find_valid_files {
  my $start_dir = $_[0];
  my $indent = $_[1];
  my @dir_contents;

  my $extensions = join ",", @VALID_EXTENSIONS;

  unless ($indent) {
    if (-t STDOUT) {
      print "Looking for valid files ($extensions):\n";
    } else {
      print "<H3>Looking for valid files ($extensions):</H3>\n";
    }
  }

  $start_dir =~ /^(.*)$/; #un-taint for running suid
  chdir ($1) or die "Cannot chdir to $start_dir: $!\n";
  
  opendir (DIR, $start_dir) or die "Cannot open $start_dir: $!";
  @dir_contents = grep {not -l} readdir (DIR);
  closedir (DIR);

  @dir_contents = sort {if (-d $a and -f $b) {return 1} else {return 0}} @dir_contents;

  if (-t STDOUT) {
    print $tab x $indent; print "+ $start_dir\n";
  } else {
    print "<DL>\n";
    print "<DD><FONT><U>$start_dir</U></FONT></DD>\n";
  }

  foreach (@dir_contents) {
    $start_dir =~ /^(.*)$/; #un-taint for running suid
    chdir ($1);
    if (-f $_) { #file
      /\.(.*)$/;
      if ($1 and member ($1, \@VALID_EXTENSIONS)) {
	if (member ($start_dir.$_, \@exclude_files)) { #invalid file
	  if (-t STDOUT) {
	    print $tab x ($indent + 1); print "- $_\n";
	  } else {
	    print "<DD><FONT color=\"#990000\">$_</FONT></DD>\n";
	  }
	} else { #valid file
	  $f_index++;
	  $file_index{$f_index}{file}=$start_dir.$_;
	   
	  $index{$f_index}{meta_d}={};
	  $index{$f_index}{meta_k}={};
	  $index{$f_index}{title}={};
	  $index{$f_index}{content}={};
	
	  if (-t STDOUT) {
	    print $tab x ($indent + 1); print "+ $_ \n";
	  } else {
	    print "<DD>$_</DD>\n";
	  }
	}
      }
    } elsif (not /^\.$/ and not /^\.\.$/ and -d $_) { #dir
      if (member ($start_dir.$_."/", \@exclude_dirs)) { #invalid dir
	if (-t STDOUT) {
	  print $tab x ($indent + 1); print "- ",$start_dir,$_,"/\n";
	} else {
	  print "<DL><DD><FONT color=\"#990000\"><U>$start_dir$_/</U></FONT></DD></DL>\n";
	}
      } else { #valid dir
	  find_valid_files ($start_dir.$_."/", $indent+1);
	}
    }
  }
  unless (-t STDOUT) {
    print "</DL>\n";
  }
}

sub load_exclude_files {
  if (-t STDOUT) {
    print "Loading excluded files:\n";
  } else {
    print "<H3>Loading excluded files:</H3>\n";
    print "<OL>\n";
  }
  if (-e $EXCLUDE_FILES_FILENAME) {
    open (FILE, $EXCLUDE_FILES_FILENAME) or next;
    while (<FILE>) {
      chomp;
      unless (/^\s*\#/ or /^\s*$/) {
	if (-e $_) {
	  push @exclude_files, $_;
	  if (-t STDOUT) {
	    print "\t$_\n";
	  } else {
	    print "<LI>$_</LI>\n";
	  }
	} else {
	  if (-t STDOUT) {
	    print "\tWARNING: \'$_\' does not exist!\n";
	  } else {
	    print "<LI><FONT color=\"#990000\"><B>WARNING</B></FONT>: \'$_\' does not exist!</LI>\n";
	  }
	}
      }
    }
    close (FILE);
    print "\n";
  } else {
    if (-t STDOUT) {
      print "WARNING: \'$EXCLUDE_FILES_FILENAME\' file does not exist!\n\n";
    } else {
      print "</OL><FONT color=\"#990000\"><B>WARNING</B></FONT>: \'$EXCLUDE_FILES_FILENAME\' file does not exist!<P>\n";
    }
  }

  unless (-t STDOUT) {
    print "</OL>\n";
  }
}

sub load_exclude_dirs {
  if (-t STDOUT) {
    print "Loading excluded directories:\n";
  } else {
    print "<H3>Loading excluded directories:</H3>\n";
    print "<OL>\n";
  }

  if (-e $EXCLUDE_DIRS_FILENAME) {
    open (FILE, $EXCLUDE_DIRS_FILENAME) or next;
    while (<FILE>) {
      chomp;
      unless (/^\s*\#/ or /^\s*$/) {
	$_ =~ s/[\/\s]*$/\//; #remove any trailing spaces and slashes and append one slash
	if (-e $_) {
	  push @exclude_dirs, $_;
	  if (-t STDOUT) {
	    print "\t$_\n";
	  } else {
	    print "<LI>$_</LI>\n";
	  }
	} else {
	  if (-t STDOUT) {
	    print "\tWARNING: \'$_\' does not exist!\n";
	  } else {
	    print "<LI><FONT color=\"#990000\"><B>WARNING</B></FONT>: \'$_\' does not exist!</LI>\n";
	  }
	}
      }
    }
    close (FILE);
    print "\n";
  } else {
    if (-t STDOUT) {
      print "WARNING: \'$EXCLUDE_DIRS_FILENAME\' file does not exist!\n\n";
    } else {
      print "</OL><FONT color=\"#990000\"><B>WARNING</B></FONT>: \'$EXCLUDE_DIRS_FILENAME\' file does not exist!</P>\n";
    }
  }
  
  unless (-t STDOUT) {
    print "</OL>\n";
  }
}

sub load_stopwords {
  if (-t STDOUT) {
    print "Loading stopwords:\n";
  } else {
    print "<H3>Loading stopwords:</H3>\n";
    print "<OL>\n";
  }
  
  if (-e $STOPWORDS_FILENAME) {
    open (FILE, $STOPWORDS_FILENAME) or die "Cannot open $STOPWORDS_FILENAME: $!";
    while (<FILE>) {
      chomp;
      unless (/^\s*\#/ or /^\s*$/) {
	$_ =~ s/\s*//g;
	push @stopwords, lc $_;
	if (-t STDOUT) {
	  print "- ",lc $_,"\n";
	} else {
	  print "<LI>",lc $_,"</LI>\n";
	}
      }
    }
  } else {
    if (-t STDOUT) {
      print "WARNING: \'$STOPWORDS_FILENAME\' file does not exist!\n\n";
    } else {
      print "</OL><FONT color=\"#990000\"><B>WARNING</B></FONT>: \'$STOPWORDS_FILENAME\' file does not exist!<P>\n";
    }
  }
 
  unless (-t STDOUT) {
    print "</OL>\n";
  }
}
sub create_index {
  my $buffer;
  my $f_index;
  my $file;
  my @valid_files;

  @valid_files=keys %index;
  
  unless (-t STDOUT) {
    print "<H3>Indexing:</H3>\n";
    print "<OL>\n";
  }

  foreach $f_index (sort {$a<=>$b} @valid_files) {
    undef $buffer;
    $file=$file_index{$f_index}{file};

    open (VFILE, $file) or next;
    while(<VFILE>) {
      $buffer.=$_;
    }
    close(VFILE);

    if (-t STDOUT) {
      print "Indexing $file_index{$f_index}{file} ($f_index)\n";
    } else {
      print "<LI>$file_index{$f_index}{file}</LI>\n";
    }
    
    count_meta_description($buffer, $f_index);
    count_meta_keywords($buffer, $f_index);
    count_title($buffer, $f_index);
    count_content($buffer, $f_index);
  }

  unless (-t STDOUT) {
    print "</OL>\n";
  }
}
  
sub count_meta_description {
  my $f_index;
  my $buffer;
  my $description;
  my $word;
  my @words;
  my %frequency;

  $buffer=$_[0];
  $f_index=$_[1];

  #print "Counting inside meta description tag in file: $file\n";

  if ($buffer=~/<meta\s+name\s*=\s*[\"\']{0,1}description[\"\']{0,1}\s+content\s*=\s*[\"\'](.*?)[\"\']>/is) {
    $description=$1;
    $file_index{$f_index}{description}=$description;
    $description=&normalize($description);
    @words=split " ",$description;
    #print "words: @words\n";
    foreach (@words) {
      #print "$_\n";
      if ($_) {
	($index{$f_index}{meta_d}{$_})++;
      }
    }
  } 
}

sub count_meta_keywords {
  my $f_index;
  my $buffer;
  my $keywords;
  my $word;
  my @words;
  my %frequency;

  $buffer=$_[0];
  $f_index=$_[1];

  #print "Counting inside meta keywords tag in file: ",$file_index{$f_index}{file},"\n";

  if ($buffer=~/<meta\s+name\s*=\s*[\"\']{0,1}keywords[\"\']{0,1}\s+content\s*=\s*[\"\'](.*?)[\"\']>/is) {
    $keywords=$1;
    $keywords=&normalize($keywords);
    @words=split " ",$keywords;
    #print "words: @words\n";
    foreach (@words) {
      #print "$_\n";
      if ($_) {
	($index{$f_index}{meta_k}{$_})++;
      }
    }
  } 
}

sub count_title {
  my $f_index;
  my $buffer;
  my $title;
  my $word;
  my @words;
  
  $buffer=$_[0];
  $f_index=$_[1];

  #print "Counting inside title tag in file: $file\n";

  if ($buffer=~/<\s*title\s*>(.*?)<\s*\/title\s*>/is) {
    $title=$1;
    $file_index{$f_index}{title}=$title;
    $title=&normalize($title);
    @words=split " ",$title;
    #print "words: @words\n";
    foreach $word (@words) {
      #print "$word\n";
      if ($word) {
	($index{$f_index}{title}{$word})++;
      }
    }
  } 
}

sub count_content {
  my $f_index;
  my $buffer;
  my $content;
  my $word;
  my @words;
  
  $buffer=$_[0];
  $f_index=$_[1];

  #print "Counting inside content in file: $file\n";

  $buffer=~s/<.*?>//sg;
  $buffer=&normalize($buffer);
  #print $buffer;
  @words=split " ",$buffer;
  #print "words: @words\n";
  foreach $word (@words) {
    #print "$word\n";
    if ($word) {
      ($index{$f_index}{content}{$word})++;
    }
  }
}

sub normalize {
  my $text=$_[0];
   
  $text=~s/\n/ /g;
  $text=~s/\W/ /g;
  $text=~s/_/ /g;
  $text=~s/\d+/ /g;
  $text=~s/\s+/ /g;
  $text=~s/\b\w{0,$MIN_WORD_LENGTH}\b/ /g;

  foreach (@stopwords) {
    $text =~ s/\b$_\b//gi;
  }

  return lc $text;
}

sub flatten_index {
  my $file;
  my $word;

  foreach $file (keys %index) {
    foreach $word (keys %{$index{$file}{meta_d}}) {
      $freq{$word}{$file}=$META_D_WEIGHT*$index{$file}{meta_d}{$word};
    }
    foreach $word (keys %{$index{$file}{meta_k}}) {
      $freq{$word}{$file}=$META_K_WEIGHT*$index{$file}{meta_k}{$word};
    }
    foreach $word (keys %{$index{$file}{title}}) {
      $freq{$word}{$file}=$TITLE_WEIGHT*$index{$file}{title}{$word};
    }
    foreach $word (keys %{$index{$file}{content}}) {
      $freq{$word}{$file}=$CONTENT_WEIGHT*$index{$file}{content}{$word};
    }
  }  
}

sub member {
  my $arg=$_[0];
  my @ary=@{$_[1]};
    
  foreach (@ary) {
    if ($_ eq $arg) {
      return 1;
    }
  }
  return 0;
}
